#include "field.h"
#include "XMLtext.h"
#include "word.h"
#include "sgmltag.h"
#include <string.h>
#include <assert.h>
#include <stdlib.h>


static void printXML(
#if STREAM
                     ostream * fpo
#else
                     FILE * fpo
#endif
                     ,const char * s)
    {
#if STREAM
    while(*s)
        {
        switch(*s)
            {
            case '<':
                fpo << "&lt;";
                break;
            case '>':
                fpo << "&gt;";
                break;
            case '&':
                if(strchr(s+1,';'))
                    fpo << '&';
                else
                    fpo << "&amp;";
                break;
            case '\'':
                fpo << "&apos;";
                break;
            case '"':
                fpo << "&quot;";
                break;
            default:
                fpo << *s;
            }
        ++s;
        }
#else
    while(*s)
        {
        switch(*s)
            {
            case '<':
                fprintf(fpo,"&lt;");
                break;
            case '>':
                fprintf(fpo,"&gt;");
                break;
            case '&':
                if(strchr(s+1,';'))
                    fputc('&',fpo);
                else
                    fprintf(fpo,"&amp;");
                break;
            case '\'':
                fprintf(fpo,"&apos;");
                break;
            case '"':
                fprintf(fpo,"&quot;");
                break;
            default:
                fputc(*s,fpo);
            }
        ++s;
        }
#endif
    }


corpus::corpus():startOfLine(4532),endOfLine(0),numberOfTokens(0),wordno(0),Token(NULL),CurrToken(NULL)
    {
    static char STAART[] = "STAART";
    staart.setWordpos(STAART,STAART+strlen(STAART));
    staart.setPOSpos(STAART,STAART+strlen(STAART));
    staart.Pos = STAART;
    }

class crumb
    {
    private:
        char * e;
        int L;
        crumb * next;
    public:
        crumb(const char * s,int len,crumb * Crumbs):L(len),next(Crumbs)
            {
            e = new char[len+1];
            strncpy(e,s,len);
            e[len] = '\0';
            }
        ~crumb()
            {
            delete [] e;
            }
        const char * itsElement()
            {
            return e;
            }
        crumb * pop(const char * until,int len)
            {
            crumb * nxt = next;
            if(L == len && !strncmp(e,until,len))
                {
                delete this;
                return nxt;
                }
            else
                {
                delete this;
                return nxt ? nxt->pop(until,len) : NULL;
                }
            }
        void print()
            {
            if(next)
                next->print();
            printf("->%s",e);
            }
        bool is(const char * elm)
            {
            return !strcmp(e,elm);
            }
    };

bool XMLtext::analyseThis()
    {
    if(element)
        {
        return Crumbs && Crumbs->is(element);
        }
    else
        {
        if(ancestor)
            {
            return Crumbs != NULL;
            }
        else
            {
            return true;
            }
        }
    }


const char * XMLtext::convert(const char * s)
    {
    const char * p = strchr(s,'&');
    if(p && strchr(p+1,';'))
        {
        static char * ret[2] = {NULL,NULL};
        static size_t len[2] = {0,0};
        static int index = 0;
        index += 1;
        index %= 2;
        if(len[index] <= strlen(s))
            {
            delete ret[index];
            len[index] = strlen(s)+1;
            ret[index] = new char[len[index]];
            }
        char *p = ret[index];
        while(*s)
            {
            if(*s == '&')
                {
                ++s;
                if(!strncmp(s,"lt;",3))
                    {
                    *p++ = '<';
                    s += 3;
                    }
                else if(!strncmp(s,"gt;",3))
                    {
                    *p++ = '>';
                    s += 3;
                    }
                else if(!strncmp(s,"amp;",4))
                    {
                    *p++ = '&';
                    s += 4;
                    }
                else if(!strncmp(s,"apos;",5)) // Notice: XML standard, but not HTML
                    {
                    *p++ = '\'';
                    s += 5;
                    }
                else if(!strncmp(s,"quot;",5))
                    {
                    *p++ = '"';
                    s += 5;
                    }
                else
                    *p++ = '&';
                }
            else
                *p++ = *s++;
            }
        *p = '\0';
        return ret[index];
        }
    return s;
    }


void XMLtext::CallBackStartElementName()
    {
    startElement = ch;
    }

void XMLtext::CallBackEndElementName()
    {
    endElement = ch;
    //printf("element(%s):%.*s\n",ClosingTag ? "closing" : "",endElement - startElement,startElement);
    if(ClosingTag)
        {
        if(Crumbs)
            {
            Crumbs = Crumbs->pop(startElement,endElement - startElement);
            if(!Crumbs)
                { // End of segment reached
                //printf("!Crumbs %d\n",this->total);
                if(this->total > 0 && this->Token)
                    {
                    this->Token[total-1].lastOfLine = true;
                    this->Token[total].firstOfLine = true;
/*                    printf("%d %*s\n"
                        ,this->total
                        ,this->Token[total-1].WordGetEnd() - this->Token[total-1].WordGetStart()
                        ,this->Token[total-1].WordGetStart()
                        );*/
                    }
                }
            }
        ClosingTag = false;
        }
    else
        {
        if(  Crumbs
          || !this->ancestor
          ||    startElement + strlen(this->ancestor) == endElement 
             && !strncmp(startElement,this->ancestor,endElement - startElement)
          )
            {
            Crumbs = new crumb(startElement,endElement - startElement,Crumbs);
            /* Bart 20091006 */
            if(  segment 
              && this->total > 0 
              && this->Token 
              && !strncmp(startElement,segment,endElement - startElement)
              )
                {
                this->Token[total-1].lastOfLine = true;
                this->Token[total].firstOfLine = true;
                }
            }
        }
/*    if(Crumbs)
        {
        Crumbs->print();
        printf("\n");
        }*/
    }

void XMLtext::CallBackStartAttributeName()
    {
    if(analyseThis())
        {
        ClosingTag = false;
        startAttributeName = ch;
        }
    }

void XMLtext::CallBackEndAttributeNameCounting()
    {
    if(analyseThis())
        {
        endAttributeName = ch;
        if(  this->wordAttributeLen == endAttributeName - startAttributeName 
            && !strncmp(startAttributeName,this->wordAttribute,wordAttributeLen )
            )
            {
            ++total;
            }
        }
    }

void XMLtext::CallBackEndAttributeNameInserting()
    {
    if(analyseThis())
        {
        endAttributeName = ch;
//        printf("attribu:%.*s\n",endAttributeName - startAttributeName,startAttributeName);

        WordPosComing = false;
        PreTagPosComing = false;
        POSPosComing = false;


        if(  this->wordAttributeLen == endAttributeName - startAttributeName 
            && !strncmp(startAttributeName,this->wordAttribute,wordAttributeLen )
            )
            {
            WordPosComing = true;
            }
        else if(  this->POSAttributeLen == endAttributeName - startAttributeName 
            && !strncmp(startAttributeName,this->POSAttribute,POSAttributeLen)
            )
            {
            POSPosComing = true;
            }
        else if(  this->PreTagAttributeLen == endAttributeName - startAttributeName 
            && !strncmp(startAttributeName,this->PreTagAttribute,PreTagAttributeLen )
            )
            {
            PreTagPosComing = true;
            }
        }
    }

void XMLtext::CallBackStartValue()
    {
    if(analyseThis())
        {
        startValue = ch;
        }
    }

void XMLtext::CallBackEndValue()
    {
    if(analyseThis())
        {
        endValue = ch;            
//        printf("value  :%.*s\n",endValue - startValue,startValue);
        if(WordPosComing)
            {
            this->Token[total].setWord(startValue,endValue,this);
            }
        else if(PreTagPosComing)
            {
            this->Token[total].setPreTagpos(startValue,endValue);
            }
        else if(POSPosComing)
            {
            this->Token[total].setPOSpos(startValue,endValue);
            }
        }
    }

void XMLtext::CallBackNoMoreAttributes()
    {
    if(analyseThis())
        {
        //token * Tok = getCurrentToken();//Token + total;
        token * Tok = CurrToken;
        if(Tok && Tok->WordGetStart() != NULL && Tok->WordGetEnd() != Tok->WordGetStart())
            { // Word as attribute
            char * EW = Tok->WordGetEnd();
            char savW = *EW;
            *EW = '\0';
            if(Tok->PreTagGetStart() != NULL)
                {
                char * EP = Tok->PreTagGetEnd();
                char savP = *EP;
                *EP = '\0';
                const char * pretag = convert(Tok->PreTagGetStart());
                Tok->Pos = Tok->PreTag = new char[strlen(pretag)+1];
                strcpy(Tok->PreTag,pretag);
                *EP = savP;
                }
            //Tok->wordPrePos = insert(Tok->WordGetStart());
            *EW = savW;
            CurrToken = NULL;
            }
        }
    }

void XMLtext::CallBackEndTag()
    {
    ClosingTag = true;
    }

void XMLtext::CallBackEmptyTag()
    {
    if(Crumbs)
        Crumbs = Crumbs->pop(startElement,endElement - startElement);
    /*if(Crumbs)
        {
        Crumbs->print();
        printf("\n");
        }*/
    }
/*
class readValue
    {
    private:
        char * word;
        size_t len;
        int pos;
    public:
        const char * isA() //20080213
            {
            return "Value";
            }
        readValue()
            {
            len = 20;
            word = new char[len];
            word[0] = '\0';
            pos = 0;
            }
        ~readValue()
            {
            delete word;
            }
        void reset()
            {
            pos = 0;
            }
#ifndef CONSTSTRCHR
        const 
#endif
        char * getString()
            {
//            printf("%s\n",word);
            if(pos == 0)
                return "";
//                word[pos] = '\0';
            return word;
            }
        char * read(char * kar)
            {
            char * nxt = kar;
            if(*nxt == ' ' || *nxt == '\t' || *nxt == '\n ' || *nxt == '\r')
                nxt = NULL;
            if(nxt && *nxt)
                {
                size_t ln = strlen(nxt);
                if(pos + ln >= len)
                    {
                    len = pos + ln + 1;
                    char * nw = new char[len];
                    strcpy(nw,word);
                    delete word;
                    word = nw;
                    }
                strcpy(word + pos,nxt);
                pos += ln;
                assert(strlen(word) < len);
                }
            return NULL;
            }
    };
*/

class wordReader
    {
    private:
        field * format;
        field * nextfield;
        field * wordfield;
        unsigned long newlines;
        unsigned long lineno;
        const char * tag;
        int kar;
        int lastkar;
        XMLtext * Text;
        char kars[2];
//        readValue * wordfield;
    public:
        unsigned long getNewlines()
            {
            return newlines;
            }
        /*
        const char * getTag()
            {
            return tag;
            }
            */
        unsigned long getLineno()
            {
            return lineno;
            }
        void initWord()
            {
            nextfield = format;
            newlines = 0;
            format->reset();
            kars[1] = '\0';
            if(lastkar)
                {
                if(lastkar == '\n')
                    ++newlines;
                kars[0] = (char)lastkar;
                lastkar = 0;
                nextfield->read(kars,nextfield);
                }
            token * Tok = Text->getCurrentToken();
            if(Tok)
                {
                if(kars[0])
                    Tok->setWordpos(Text->ch-1,Text->ch-1);
                else
                    Tok->setWordpos(Text->ch,Text->ch);
                }
            }
/*
        void initWord()
            {
            newlines = 0;
            kars[1] = '\0';
            if(lastkar)
                {
                if(lastkar == '\n')
                    ++newlines;
                kars[0] = (char)lastkar;
                lastkar = 0;
                }
            token * Tok = Text->getCurrentToken();
            if(Tok)
                {
                if(kars[0])
                    Tok->setWordpos(Text->ch-1,Text->ch-1);
                else
                    Tok->setWordpos(Text->ch,Text->ch);
                }
            wordfield = new readValue();
            }
        ~wordReader()
            {
            delete wordfield;
            wordfield = NULL;
            }

        wordReader(XMLtext * Text)
            :newlines(0)
            ,lineno(0)
            ,tag(NULL)
            ,kar(EOF)
            ,lastkar(0)
            ,Text(Text)
            {
            initWord();
            if(lastkar)
                {
                assert(lastkar != EOF);
                if(lastkar == '\n')
                    ++newlines;
                kars[0] = (char)lastkar;
                lastkar = 0;
                }
            }*/
        wordReader(field * format,field * wordfield/*,field * tagfield,bool treatSlashAsAlternativesSeparator*/,XMLtext * Text)
            :format(format)
            ,nextfield(format)
            ,wordfield(wordfield)
            //,tagfield(tagfield)
            ,newlines(0)
            ,lineno(0)
            ,tag(NULL)
            ,kar(EOF)
            ,lastkar(0)
            //,treatSlashAsAlternativesSeparator(treatSlashAsAlternativesSeparator)
            ,Text(Text)
            {
            initWord();
            assert(wordfield);
            if(lastkar)
                {
                assert(lastkar != EOF);
                if(lastkar == '\n')
                    ++newlines;
                kars[0] = (char)lastkar;
                lastkar = 0;
                nextfield->read(kars,nextfield);
                }
            }
#ifndef CONSTSTRCHR
        const 
#endif
        char * readChar(int kar)
            {
            if(!nextfield)
                {
                initWord();
                }
            if(kar == '\n')
                ++newlines;
            kars[0] = kar == EOF ? '\0' : (char)kar;
            char * plastkar = nextfield->read(kars,nextfield);
            if(kar == '\0')
                lastkar = '\0';
            else if(plastkar)
                lastkar = *plastkar;

            if(nextfield)
                return NULL;
            else
                {
                lineno += newlines;
                /*if(tagfield)
                    tag = tagfield->getString();*/
                return wordfield->getString();
                }
            }
        /*
        char * readChar(int kar)
            {
            //if(!nextfield)
                {
                initWord();
                }
            if(kar == '\n')
                ++newlines;
            kars[0] = kar == EOF ? '\0' : (char)kar;
            char * plastkar = wordfield->read(kars);
            if(kar == '\0')
                lastkar = '\0';
            else if(plastkar)
                lastkar = *plastkar;

            lineno += newlines;
            return wordfield->getString();
            }
            */
#ifndef CONSTSTRCHR
        const 
#endif
        char * count(int kar)
            {
#ifndef CONSTSTRCHR
            const 
#endif
            char * w = readChar(kar);
            if(Text->analyseThis())
                {
                if(/*!Text->wordAttribute &&*/ w && *w)
                    {
                    Text->incTotal();
                    }
                }
            return w;
            }
#ifndef CONSTSTRCHR
        const 
#endif
        char * read(int kar)
            {
#ifndef CONSTSTRCHR
            const 
#endif
            char * w = readChar(kar);
            if(Text->analyseThis())
                {
                if(w && *w)
                    {
                    token * Tok = Text->getCurrentToken();
                    char * start = Tok->WordGetStart();
                    Tok->setWord(start,start+strlen(w),Text);
                    const char * pretag = Tok->PreTagGetStart();
                    if(pretag != NULL)
                        {
                        char * EP = Tok->PreTagGetEnd();
                        char savP = *EP;
                        *EP = '\0';
                        pretag = Text->convert(pretag);
                        Tok->Pos = Tok->PreTag = new char[strlen(pretag)+1];
                        strcpy(Tok->PreTag,pretag);
                        *EP = savP;
                        char temp[100];
                        sprintf(temp,"[%s]",Tok->Pos);
                        }
                    Text->CurrToken = NULL;
                    /*char * POS = Tok->POS.getStart();
                    if(POS != NULL)
                        {
                        char * EP = Tok->POS.getEnd();
                        char savP = *EP;
                        *EP = '\0';
                        Text->createTagged(w,POS);
                        *EP = savP;
                        }
                    else if(tag)
                        {
                        Text->createTagged(w,tag);
                        }
                    else*/
                        {
//                        Tok->wordPrePos = Text->createUnTagged(w);
                        }
                    }
                }
            return w;
            }
    };

void XMLtext::printUnsorted(
#if STREAM
                            ostream & fpo
#else
                            FILE * fpo            
#endif
                            )
    {
    if(this->alltext)
        {
        unsigned int k;
        const char * o = this->alltext;
        for(k = 0;k < total;++k)
            {
            //if(tunsorted[k])
                {
                const char * start = Token[k].WordGetStart();
                const char * end = Token[k].WordGetEnd();
                const char * POSstart = Token[k].POSGetStart();
                const char * POSend = Token[k].POSGetEnd();
                char * p = strchr(Token[k].Pos,' ');
                if(p)
                    *p = 0;
                if(POSstart == NULL)
                    { // write word together with POS-tag
                    if(start)
                        {
                        copy(fpo,o,end);
                        o = end;
                        }
                    //tunsorted[k]->print();
                    //TODO:print computed PoS-tag. TOKEN-WISE! fprintf(fpo,"%s",tunsorted[k]->m_tag);
                    printXML(fpo,"/");
                    printXML(fpo,Token[k].Pos);
                    }
                else
                    {
                    copy(fpo,o,POSstart);
                    o = POSend;
                    //tunsorted[k]->print();
                    //TODO:print computed PoS-tag. TOKEN-WISE! fprintf(fpo,"%s",tunsorted[k]->m_tag);
                    printXML(fpo,Token[k].Pos);
                    }
                if(p)
                    *p = ' ';

                /*const char * POSstart = Token[k].POS.getStart();
                const char * POSend = Token[k].POS.getEnd();
                copy(fpo,o,POSstart);
                o = POSend;
                fprintf(fpo,"%s",tunsorted[k]->m_tag);
                */
                }
            }
        o = copy(fpo,o,o+strlen(o));
        }
    }

token * XMLtext::getCurrentToken()
    {
    //return CurrToken;
    return Token ? Token + total : NULL;
    }

void CallBackStartElementName(void *arg)
    {
    ((XMLtext *)arg)->CallBackStartElementName();
    }

void CallBackEndElementName(void *arg)
    {
    ((XMLtext *)arg)->CallBackEndElementName();
    }

void CallBackStartAttributeName(void *arg)
    {
    ((XMLtext *)arg)->CallBackStartAttributeName();
    }

static void (XMLtext::*CallBackEndAttributeNames)();

void CallBackEndAttributeName(void *arg)
    {
    (((XMLtext *)arg)->*CallBackEndAttributeNames)();
    }

void CallBackStartValue(void *arg)
    {
    ((XMLtext *)arg)->CallBackStartValue();
    }

void CallBackEndValue(void *arg)
    {
    ((XMLtext *)arg)->CallBackEndValue();
    }

void CallBackNoMoreAttributes(void *arg)
    {
    ((XMLtext *)arg)->CallBackNoMoreAttributes();
    }

void CallBackEndTag(void * arg)
    {
    ((XMLtext *)arg)->CallBackEndTag();
    }

void CallBackEmptyTag(void * arg)
    {
    ((XMLtext *)arg)->CallBackEmptyTag();
    }

#if STREAM
XMLtext::XMLtext(istream & fpi
           ,char * Iformat
           ,bool nice
           ,unsigned long int size
           ,bool XML
           ,const char * ancestor // if not null, restrict POS-tagging to elements that are offspring of ancestor
           ,const char * segment // Optional segment delimiter. E.g. br: <br /> or s: <s> ... </s>
           ,const char * element // if null, analyse all PCDATA that is text
           ,const char * wordAttribute // if null, word is PCDATA
           ,const char * PreTagAttribute // Fetch pretagging from PreTagAttribute
           ,const char * POSAttribute // Store POS in POSAttribute
           )
#else
XMLtext::XMLtext(FILE * fpi
           //,bool a_InputHasTags
           ,char * Iformat
           //,int keepPunctuation
           ,bool nice
           ,unsigned long int size
           //,bool treatSlashAsAlternativesSeparator
           ,bool XML
           ,const char * ancestor // restrict POS-tagging to segments that fit in ancestor elements
           ,const char * segment // Optional segment delimiter. E.g. br: <br /> or s: <s> ... </s>
           ,const char * element // analyse PCDATA inside elements
           ,const char * wordAttribute // if null, word is PCDATA
           ,const char * PreTagAttribute // Fetch pretagging from PreTagAttribute
           ,const char * POSAttribute // Store POS in POSAttribute
           )
#endif
           :text(nice)
           ,ancestor(ancestor)
           ,segment(segment)
           ,element(element)
           ,PreTagAttribute(PreTagAttribute)
           ,POSAttribute(POSAttribute)
           ,Crumbs(NULL)
           ,ClosingTag(false)
           ,WordPosComing(false)
           ,PreTagPosComing(false)
           ,POSPosComing(false)
           ,wordAttribute(wordAttribute)
    {
    wordAttributeLen = wordAttribute ? strlen(wordAttribute) : 0;
    PreTagAttributeLen = PreTagAttribute ? strlen(PreTagAttribute) : 0;
    POSAttributeLen = POSAttribute ? strlen(POSAttribute) : 0;
#ifdef COUNTOBJECTS
    ++COUNT;
#endif

    field * wordfield;
    //field * tagfield;
    field * format = 0;
    //if(nice)
      //  printf("counting words\n");
//    lineno = 0;
    
    if(XML && Iformat)
        {
#if STREAM
        fpi.seekg(0,ios::end);
        long filesize = fpi.tellg();
        fpi.seekg(0,ios::beg);
        if(filesize > 0)
            {
            alltext = new char[filesize+1];
            char * p = alltext;
            while(fpi.good())
                {
                fpi >> *p++;
                }
            *p = '\0';
            fpi.seekg(0,ios::beg);
            }
#else
        int kar;
        fseek(fpi,0,SEEK_END);
        long filesize = ftell(fpi);
        ::rewind(fpi);
        if(filesize > 0)
            {
            alltext = new char[filesize+1];
            char * p = alltext;
            while((kar = getc(fpi)) != EOF)
                *p++ = (char)kar;
            *p = '\0';
            ::rewind(fpi);
            }
#endif
        }
    if(alltext)
        {
        html_tag_class html_tag(this);
        CallBackEndAttributeNames = &XMLtext::CallBackEndAttributeNameCounting;
        if(element||ancestor)
            {
            html_tag.setCallBackStartElementName(::CallBackStartElementName);
            html_tag.setCallBackEndElementName(::CallBackEndElementName);
            html_tag.setCallBackEndTag(::CallBackEndTag);
            html_tag.setCallBackEmptyTag(::CallBackEmptyTag);
            }
        else
            {
            html_tag.setCallBackStartElementName(::dummyCallBack);
            html_tag.setCallBackEndElementName(::dummyCallBack);
            html_tag.setCallBackEndTag(::dummyCallBack);
            html_tag.setCallBackEmptyTag(::dummyCallBack);
            }
        if(wordAttribute||POSAttribute||PreTagAttribute)
            {
            html_tag.setCallBackStartAttributeName(::CallBackStartAttributeName);
            html_tag.setCallBackEndAttributeName(::CallBackEndAttributeName);
//            print = printXML;
            }
        else
            {
            html_tag.setCallBackStartAttributeName(::dummyCallBack);
            html_tag.setCallBackEndAttributeName(::dummyCallBack);
            }
        html_tag.setCallBackStartValue(::dummyCallBack);
        html_tag.setCallBackEndValue(::dummyCallBack);
        html_tag.setCallBackNoMoreAttributes(::dummyCallBack);
        ch = alltext;
        char * curr_pos = alltext;
        char * endpos = alltext;
        estate Seq = notag;
#ifndef CONSTSTRCHR
        const 
#endif
        char * (wordReader::*fnc)(int kar);
        int loop;
        fnc = &wordReader::count;
        if(Iformat)
            {
            wordfield = 0;
            //tagfield = 0;
            format = translateFormat(Iformat,wordfield/*,tagfield*/);
            if(!wordfield)
                {
                printf("Input format %s must specify '$w'.\n",Iformat);
                exit(0);
                }
            }
        for(loop = 1;loop <= 2;++loop)
            {
            wordReader WordReader(format,wordfield/*,tagfield,treatSlashAsAlternativesSeparator*/,this);
            //wordReader WordReader(/*treatSlashAsAlternativesSeparator,*/this);
            while(*ch)
                {
                while(  *ch 
                     && (( Seq = (html_tag.*tagState)(*ch)) == tag 
                        || Seq == endoftag_startoftag
                        )
                     )
                    {
                    ch++;
                    }
                if(Seq == notag)
                    { // Not an HTML tag. Backtrack.
                    endpos = ch;
                    ch = curr_pos;
                    //printf("NOTAG:");  // TEXT
                    while(ch < endpos)
                        {
                        (WordReader.*fnc)(*ch);
                        //putchar(*ch);  // TEXT
                        ch++;
                        }
                    }
                else if(Seq == endoftag)
                    ++ch; // skip >
                if(*ch)
                    {
//                    printf("[");  // TEXT START
                    while(  *ch 
                        && (Seq = (html_tag.*tagState)(*ch)) == notag
                        )
                        {
                        (WordReader.*fnc)(*ch);
                        //putchar(*ch); // TEXT
                        ++ch;
                        }
                    (WordReader.*fnc)('\0');
                    if(Seq == tag)
                        {
                        (WordReader.*fnc)('\0');
                        curr_pos = ch++; // skip <
                        //printf("]");  // TEXT END
                        }
                    }
                }
            if(Seq == tag)
                { // Not an SGML tag. Backtrack.
                endpos = ch;
                ch = curr_pos;
                //printf("NOTAG:");  // TEXT
                while(ch < endpos)
                    {
                    (WordReader.*fnc)(*ch);
                    //putchar(*ch); // TEXT
                    ch++;
                    }
                }
            if(loop == 1)
                {
//                lineno = WordReader.getLineno();
                /*if(nice)
                    {
                    printf("... %lu words counted in %lu lines\n",total,lineno);
                    printf("allocating array of pointers to words\n");
                    }*/
//                tunsorted =  new const Word * [total];
                numberOfTokens = total;
                Token = new token[total + 1]; // 1 extra for the epilogue after the last token
                /*if(nice)
                    printf("allocating array of line offsets\n");*/
/*                Lines =  new unsigned long int [lineno+1];
                for(int L = lineno+1;--L >= 0;)
                    Lines[L] = 0;*/
                /*if(nice)
                    printf("...allocated array\n");*/

                total = 0;
                /*if(nice)
                    printf("reading words\n");*/
//                lineno = 0;
                ch = alltext;
                curr_pos = alltext; // 20091106
                fnc = &wordReader::read;
                CallBackEndAttributeNames = &XMLtext::CallBackEndAttributeNameInserting;
                if(wordAttribute||PreTagAttribute||POSAttribute)
                    {
                    html_tag.setCallBackStartAttributeName(::CallBackStartAttributeName);
                    html_tag.setCallBackEndAttributeName(::CallBackEndAttributeName);
                    html_tag.setCallBackStartValue(::CallBackStartValue);
                    html_tag.setCallBackEndValue(::CallBackEndValue);
                    html_tag.setCallBackNoMoreAttributes(::CallBackNoMoreAttributes);
                    }
                }
            }
        }
//    makeList();
    /*if(nice)
        printf("...read words from XML file\n");*/
//    sorted = false;
    }

